/*
 ATmega GPS SMS Alerter
 Version: 0.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 9/10/2013
 Last Modified: 25/01/2014
 
 Waits until the car is moving which is detected by the accelerometer, turns on the GPS and wait until it locks,
 reads the GPS Longitude and Latitude and calculates how far it is away from the alert area. If it's in the alert
 area, it sends an SMS. If it's found to be more than 5KM away from the alert area it switches off for 5 minutes but
 if it's within 5KM, the GPS stays on monitoring and if no motion is detected over 1 minute we switch the GPS off and
 check again in 1 minute
 
 Remember to change your fuse bits so the ATmega328 uses a 12MHz crystal:
 avrdude -p atmega328p -c usbtiny -U lfuse:w:0xff:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m
 
 */

// ATmega328 Pin Map
//
//               +-\/-+
//  Reset  PC6  1|o   |28  PC5 N Mosfet for MMA7361 accelerometer
//  RX GPS PD0  2|    |27  PC4 N Mosfet for GPS
//  TX SMS PD1  3|    |26  PC3 N Mosfet for Nokia 3120
//         PD2  4|    |25  PC2 X Axis
//         PD3  5|    |24  PC1 Y Axis
//         PD4  6|    |23  PC0 Z Axis
//         VCC  7|    |22  GND
//         GND  8|    |21  AREF
//    XTAL PB6  9|    |20  AVCC
//    XTAL PB7 10|    |19  PB5 SCK
//         PD5 11|    |18  PB4 MISO
//         PD6 12|    |17  PB3 MOSI
//         PD7 13|    |16  PB2 Optoisolator for Nokia 3120 power button
//         PB0 14|    |15  PB1 LED
//               +----+

#define F_CPU 12000000 // 12 MHz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <util/delay.h>
#include "setup.c"

#define X_AXIS 2 // ADC2
#define Y_AXIS 1 // ADC1
#define Z_AXIS 0 // ADC0

uint8_t tempBuffer[15];
uint8_t gpsLatitude[15];
uint8_t gpsLongitude[15];
double gpsLatitudeDec = 0;
double gpsLongitudeDec = 0;

// Change these to be a box around your alert area
double compareLatitudeUp = 41.133; // Actually a negative number
double compareLatitudeDown = 41.158; // Actually a negative number
double compareLongitudeLeft = 147.763;
double compareLongitudeRight = 147.814;

bool monitorGPS = false;
int gpsLowestDistance = 100;
volatile int eightSecondCounter = 0;

// Receive USART data
uint8_t USART_Receive( void ) {
	while ( !(UCSR0A & (1<<RXC0)) ); // Wait for data to be received
	return UDR0; // Get and return received data from buffer
}

// FbusFrame buffer
unsigned char FBusFrame[60];

// SMS Center number (your mobile phone provider should provide this)
unsigned char SMSC[] = {0x07, 0x91, 0x16, 0x14, 0x91, 0x09, 0x00, 0xF1, 0x00, 0x00, 0x00, 0x00};

// Recipient SMS Number. After the 0x81, you insert the phone number with the numbers flipped
// E.g to insert 0412345678 you would write it as below, 0x40 becomes 04, 0x21 becomes 12, etc
unsigned char RecipientNo[] = {0x0A, 0x81, 0x40, 0x21, 0x43, 0x65, 0x87};

// Transmit USART data
void USART_Transmit(unsigned char data) {
	while ( !( UCSR0A & (1<<UDRE0)) ); // Wait for empty transmit buffer
	UDR0 = data;
}

// Send an SMS
void SendSMS(const char *Message) {
	// Clear buffer
	unsigned char j = 0;
	memset(FBusFrame, 0, sizeof(FBusFrame));

	unsigned char MsgLen = strlen(Message), FrameSize = 0;
	unsigned char c, w, n, shift = 0, frameIndex = 0;
	unsigned char oddCheckSum = 0, evenCheckSum = 0, SeqNo = 0x43;
	unsigned char MsgStartIndex = 48; // Message will always start here

	// Encode the message into 7 bit characters
	for (n = 0; n < MsgLen; n++) {
		c = Message[n] & 0x7f;
		c >>= shift;
		w = Message[n+1] & 0x7f;
		w <<= (7-shift);
		shift += 1;
		c = c | w;
		if (shift == 7) {
			shift = 0x00;
			n++;
		}
		FBusFrame[frameIndex + MsgStartIndex] = c;
		frameIndex++;
	}

	FBusFrame[frameIndex + MsgStartIndex] = 0x01;
	FrameSize = frameIndex + 44; // The size of the frame is frameIndex+48 (FrameIndex + 48 + 1 - 5)

	// Insert the frame values to prepare to send an SMS 
	FBusFrame[0] = 0x1E;
	FBusFrame[1] = 0x00;
	FBusFrame[2] = 0x0C;
	FBusFrame[3] = 0x02;
	FBusFrame[4] = 0x00;
	FBusFrame[5] = FrameSize;
	FBusFrame[6] = 0x00;
	FBusFrame[7] = 0x01;
	FBusFrame[8] = 0x00;
	FBusFrame[9] = 0x01;
	FBusFrame[10] = 0x02;
	FBusFrame[11] = 0x00;

	// Insert the SMS Center number
	for (j = 0; j < sizeof(SMSC); j++) {
		FBusFrame[12 + j] = SMSC[j];
	}

	FBusFrame[24] = 0x15; //Message type??
	FBusFrame[28] = MsgLen; //Message length (uncompressed)

	// Insert the Recipient number
	for (j = 0; j < sizeof(RecipientNo); j++) {
		FBusFrame[j + 29] = RecipientNo[j];
	}

	FBusFrame[41] = 0xA7; // Validity period

	// Check if the Framesize is odd or even
	if (FrameSize & 0x01) {
		frameIndex = FrameSize + 5;
		FBusFrame[frameIndex] = SeqNo;
		frameIndex++;
		FBusFrame[frameIndex] = 0; // Insert to make the Frame even
		frameIndex++;
	}
	else {
		frameIndex = FrameSize + 5;
		FBusFrame[frameIndex] = SeqNo;
		frameIndex++;
	}

	// Calculate the checksum from the start of the frame to the end of the frame
	for (unsigned char i = 0; i < frameIndex+2; i += 2) {
		oddCheckSum ^= FBusFrame[i];
		evenCheckSum ^= FBusFrame[i+1];
	}
	FBusFrame[frameIndex] = oddCheckSum;
	FBusFrame[frameIndex+1] = evenCheckSum;

	// Send the full frame to the phone
	for (unsigned char j = 0; j < (frameIndex+2); j++) {
		USART_Transmit(FBusFrame [j]);
	}
}

// Convert the GPS NMEA to decimal (without using atof as it takes up too much space)
double gps_nmea_to_double(uint8_t* gpsData) {
	// Read until decimal place
	int x = 0;
	while (gpsData[x] != '.') {
		x++;
	}
	
	// Take the minutes from the part before the decimal point
	char tempint[5];
	memcpy(tempint, gpsData+(x-2), 2);
	tempint[3] = 0;
	int minutes = atoi((char*) &tempint); 
	
	// Take the degrees from the part before the decimal point
	memcpy(tempint, gpsData, (x-2));
	tempint[(x-2)] = 0;
	int degrees = atoi((char*) &tempint); 
	
	// Convert the part after the decimal point to an int
	uint32_t afterDecimalpoint = atol((char*) &gpsData[++x]); 
	
	// Now do the NMEA to decimal conversion
	double calculateDecimal = (double) (((double) minutes + (double) afterDecimalpoint / (double) 100000) / (double) 60) + degrees;
	
	return calculateDecimal;
}

// Calculate the distance between two GPS points and assign the shortest difference to the gpsLowestDistance varaiable. 
// 1KM is approximately 0.01 in longitude or latitude (we times by 100 to give us a round number for each KM)
void calculate_shortest_distance(double compare, double compared_to) {
	int distance = (double) (compare * 100) - (double) (compared_to * 100);
	if (distance < 0) {
		distance = distance - (distance * 2);
	}
	if (distance < gpsLowestDistance) {
		gpsLowestDistance = distance;
	}
}

// Read USART until we reach a comma
uint8_t read_usart_to_comma(void) {
	int x = 0;
	tempBuffer[x] = USART_Receive();
	
	while (tempBuffer[x] != ',') {
		x++;
		tempBuffer[x] = USART_Receive();
	}
	
	return x;
}

// Check for any movement on the accelerometer
uint8_t check_for_movement(uint8_t time_to_monitor) {
	// Turn on accelerometer and ADC
	PORTC |= (1<<MMA_MOSFET);
	sbi(ADCSRA, ADEN);
	adc_read(X_AXIS); // Ignore first reading
	watchdog_sleep(T64MS);
	
	// Measure if motion is detected over 10 seconds
	int xResult = adc_read(X_AXIS);
	int yResult = adc_read(Y_AXIS);
	int zResult = adc_read(Z_AXIS);
	
	// Read X, Y, Z axis every 250ms
	int highestValue = 0;
	for (int i = 0; i < (time_to_monitor * 4); i++) {
		int xChange = adc_read(X_AXIS) - xResult;
		int yChange = adc_read(Y_AXIS) - yResult;
		int zChange = adc_read(Z_AXIS) - zResult;
		
		// Convert negative to positive value
		if (xChange < 0) {
			xChange = xChange - (xChange * 2);
		}
		if (yChange < 0) {
			yChange = yChange - (yChange * 2);
		}
		if (zChange < 0) {
			zChange = zChange - (zChange * 2);
		}
		
		// Record which one had the highest change
		if (xChange > highestValue) {
			highestValue = xChange;
		}
		if (yChange > highestValue) {
			highestValue = yChange;
		}
		if (zChange > highestValue) {
			highestValue = zChange;
		}
		
		watchdog_sleep(T250MS);
		
		// Movement detected? 0.1g is 80mV assuming 800mV/g typical.
		if (highestValue >= 200) {
			// Turn off accelerometer and ADC
			PORTC &= ~(1<<MMA_MOSFET);
			cbi(ADCSRA, ADEN);
			
			return 1;
		}
	}
	
	// Turn off accelerometer and ADC
	PORTC &= ~(1<<MMA_MOSFET);
	cbi(ADCSRA, ADEN);
	
	return 0;
}

int main(void) {
	setup();
	
	while (1) {		
		// Was motion detected over 10 seconds?
		if (check_for_movement(10) == true) {
			monitorGPS = true;
			PORTC |= (1<<GPS_MOSFET); // Turn on GPS
		}
		else { // No motion, check again in ~1 minute
			monitorGPS = false;
			eightSecondCounter = 8;
		}
		
		// If we are monitoring the GPS, stay in this loop
		while (monitorGPS == true) {
			uint8_t rxData = USART_Receive();
			
			// Start of frame
			if (rxData == '$') {
				// Read next chars
				read_usart_to_comma();
				
				// Check if GPRMC
				if (memcmp(tempBuffer, "GPRMC", 5) == 0) {
					read_usart_to_comma(); // Timestamp (ignore)
					read_usart_to_comma(); // Validity
					
					// If valid GPS data
					if (tempBuffer[0] == 'A') {
						
						// Read Lat and Long
						uint8_t bytesReadLatitude = read_usart_to_comma(); // Latitude	
						memcpy(gpsLatitude, tempBuffer, bytesReadLatitude);
						
						read_usart_to_comma(); // N or S (is ignored because I know what it's going to be, you should check this)
						
						uint8_t bytesReadLongitude = read_usart_to_comma(); // Longitude
						memcpy(gpsLongitude, tempBuffer, bytesReadLongitude);
						
						// Convert Lat and Long to decimal
						gpsLatitudeDec = gps_nmea_to_double(gpsLatitude);
						gpsLongitudeDec = gps_nmea_to_double(gpsLongitude);
						
						// Pulse
						PORTB |= (1<<PB1);
						_delay_ms(20);
						PORTB &= ~(1<<PB1);
						
						// In SMS alert area
						if (gpsLatitudeDec > compareLatitudeUp && gpsLatitudeDec < compareLatitudeDown && 
							gpsLongitudeDec > compareLongitudeLeft && gpsLongitudeDec < compareLongitudeRight) {
							
							// LED on
							PORTB |= (1<<PB1);
							
							// Set re-check timer to 1 hour
							eightSecondCounter = 450;
							
							// Turn off GPS
							PORTC &= ~(1<<GPS_MOSFET);
							monitorGPS = false;
							
							// Disable RX
							cbi(UCSR0B, RXEN0); // Receiver disable
							
							// Enable TX
							UBRR0L = 12; // 115,200 Baud rate
							sbi(UCSR0A, U2X0); // Double rate
							sbi(UCSR0B, TXEN0); // Transmitter enable
							
							// Provide power to the phone
							PORTC |= (1<<NOKIA_MOSFET);
							
							// Wait a second and then hold the power button down to start the phone
							watchdog_sleep(T1S);
							PORTB |= (1<<NOKIA_POWER_BUTTON);
							watchdog_sleep(T2S);
							PORTB &= ~(1<<NOKIA_POWER_BUTTON);
							
							// Wait for phone to boot up
							watchdog_sleep(T8S);
							watchdog_sleep(T8S);
							
							// Transmit U 128 times
							int x;
							for (x = 0; x < 128; x++) {
								USART_Transmit('U'); 
							}
							_delay_ms(50);
							
							// Send the SMS message
							SendSMS("In area");
							_delay_ms(50);
							
							// Wait a few seconds and then turn off phone
							watchdog_sleep(T8S);
							PORTC &= ~(1<<NOKIA_MOSFET);
							PORTB &= ~(1<<PB1);
							
							// Disable TX
							cbi(UCSR0B, TXEN0);
							
							// Enable RX
							UBRR0L = 155; // 9,600 Baud rate
							sbi(UCSR0A, U2X0); // Double rate
							sbi(UCSR0B, RXEN0); // Receiver enable
							
						}
						else {
							PORTB &= ~(1<<PB1);
							
							// How far are we away from alert area? Find the shortest distance of the Latitude and Longitude.
							gpsLowestDistance = 100;
							calculate_shortest_distance(compareLatitudeUp, gpsLatitudeDec);
							calculate_shortest_distance(compareLatitudeDown, gpsLatitudeDec);
							calculate_shortest_distance(compareLongitudeLeft, gpsLongitudeDec);
							calculate_shortest_distance(compareLongitudeRight, gpsLongitudeDec);
							
							// If more than 5KM away, check again in 5 minutes
							if (gpsLowestDistance > 5) {
								// Turn off GPS
								PORTC &= ~(1<<GPS_MOSFET);
								monitorGPS = false;
								
								eightSecondCounter = 38; // Check again in 5 minutes
							}
							else { // Less than 5KM away
								if (check_for_movement(60) == false) { // If we aren't moving over 1 minute, turn off GPS
									// Turn off GPS
									PORTC &= ~(1<<GPS_MOSFET);
									monitorGPS = false;
									
									eightSecondCounter = 8; // Check again in 1 minute
								}
							}
						}
					}
				}
			}
		}
		
		// Sleep until we reach 0
		while (eightSecondCounter >= 1) {
			watchdog_sleep(T8S);
			eightSecondCounter--;
		}
	}
}

ISR(WDT_vect) { }

ISR(ADC_vect) { }
